/*
   Copyright (C) 2007 - 2009 by Roberto García Carvajal <roberpot@gmail.com>
   Part of libglsws (http://magiccards.wordpress.com)

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3
   or at your option any later version.
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY.

   See the COPYING file for more details.
*/

#include <GL/gl.h>
#include <list>
#include "c_sventana.h"
#include "c_ventana.h"
#include "c_oventana.h"
#include "c_lexico.h"
#include "c_sintactico.h"

#include <iostream>

using namespace std;

/*******************************************************************************
*
* Constructor.
* Inicializa todas las variables internas.
*
*******************************************************************************/
c_sventana::c_sventana()
{
 _RBE = 0;
 _id_foco = -1;
 _oventana = 0;
 _ventana = 0;
 _c_ventana_id = 0;
 _c_oventana_id = 0;
}

/*******************************************************************************
*
* Método parsear.
* Recibe un puntero a carácter constante que apunta un nombre de fichero.
* Genera todas las ventanas a partir de un fichero.
*
*******************************************************************************/
bool c_sventana::parsear(const GLbyte * f)
{
 // Crea el analizador sintáctico, el nodo principal y comienza a generar las ventanas.
 c_sintactico analizador(f);
 nodo_ventanas * nodo = 0;
 nodo = analizador.procesar();
 nodo = generar_ventanas(nodo);
 // FIX-ME POSIBLE RESERVA DE MEMORIA NO LIBERADA!!
 return true;
}

/*******************************************************************************
*
* Método cargar.
* Recibe un entero con signo de 32 bits i y un conjunto de datos d.
* Si existe una ventana con uid i, le pasa el conjunto de datos.
*
*******************************************************************************/
void c_sventana::cargar(GLint i, c_datos d)
{
 if(_ventanas.count(i))
 {
  _ventanas[i]->cargar(d);
 }
}

/*******************************************************************************
*
* Método nventana.
* Recibe un entero con signo de 32 bits id.
* Crea una nueva ventana con uid id.
*
*******************************************************************************/
GLint c_sventana::nventana(GLint id)
{
 c_ventana * v;
 v = new c_ventana;
 _ventanas[id] = v;
 _vordenadas.push_back(id);
 return 0;
}

/*******************************************************************************
*
* Método ventana.
* Recibe un entero con signo de 32 bits id.
* Devuelve la ventana con uid id.
*
*******************************************************************************/
c_ventana * c_sventana::ventana(GLint id)
{
 return _ventanas[id];
}

/*******************************************************************************
*
* Método ci_abajo.
* Recibe un puntero a carácter constante que apunta un nombre de fichero.
* Genera todas las ventanas a partir de un fichero.
*
*******************************************************************************/
bool c_sventana::ci_abajo(GLfloat x,GLfloat y)
{
 std::pair<bool,bool> res;
 res.first = false;
 res.second = false;
 // Recorro las ventanas hasta que encuentro una en la que he pulsado.
 list<GLint>::iterator r = _vordenadas.begin(),fin = _vordenadas.end();
 for(;r != fin && !res.first;r++)
  if(_ventanas[(*r)]->estado() & V_VISIBLE)
   res = _ventanas[(*r)]->click(x,y);

 // Si he hecho click en alguna...
 if(res.first)
 {
  r--;
  // Si es móvil la coloco al frente.
  if(_ventanas[(*r)]->estado() & V_MOVIL)
  {
   _vordenadas.remove((*r));
   _vordenadas.push_front((*r));
  }
  _id_foco = (*r);
  // Si no he hecho click en un objeto, la configuro para moverla.
  if(!res.second)
  {
   _x_foco = (GLint)(x - _ventanas[(*r)]->x());
   _y_foco = (GLint)(y - _ventanas[(*r)]->y());
   _RBE |= SV_CLICKABAJO;
  }
  else
  {
   vector<c_accion> acciones = _ventanas[_id_foco]->acciones(L_ECLICK);
   cout << "TOTAL:" << acciones.size() << endl;
   for(unsigned int i = 0;i < acciones.size();i++)
   {
    c_accion accion = acciones[i];
    if(!accion.ventana)
     continue;
    if(accion.objeto)
    {
     try
     {
      for(unsigned int j = 0;j < accion.parametros.size();j++)
      {
       switch(accion.parametros[j].t_parametro)
       {
        case L_X:
        {
         _ventanas[accion.ventana]->objeto(accion.objeto)->x(accion.parametros[j].i);
        } break;
        case L_Y:
        {
         _ventanas[accion.ventana]->objeto(accion.objeto)->y(accion.parametros[j].i);
        } break;
        case L_ANCHO:
        {
         _ventanas[accion.ventana]->objeto(accion.objeto)->ancho(accion.parametros[j].i);
        } break;
        case L_ALTO:
        {
         _ventanas[accion.ventana]->objeto(accion.objeto)->alto(accion.parametros[j].i);
        } break;
        case L_IMG:
        {
         _ventanas[accion.ventana]->objeto(accion.objeto)->textura(accion.parametros[j].i);
         _ventanas[accion.ventana]->objeto(accion.objeto)->set(V_TEXTURADO);
        } break;
        case L_VISIBLE:
        {
         if (accion.parametros[j].b)
          _ventanas[accion.ventana]->objeto(accion.objeto)->set(V_VISIBLE);
         else
          _ventanas[accion.ventana]->objeto(accion.objeto)->unset(V_VISIBLE);
        } break;
        case L_COLOR:
        {
         _ventanas[accion.ventana]->objeto(accion.objeto)->colorf(accion.parametros[j].f3);
        } break;
        case L_BCOLOR:
        {
         _ventanas[accion.ventana]->objeto(accion.objeto)->colorb(accion.parametros[j].f3);
        } break;
        case L_TCOLOR:
        {
         _ventanas[accion.ventana]->objeto(accion.objeto)->colort(accion.parametros[j].f3);
        } break;
        case L_FACTORTEXTO:
        {
         _ventanas[accion.ventana]->objeto(accion.objeto)->textoTamano(accion.parametros[j].i);
        } break;
        case L_GROSORTEXTO:
        {
         _ventanas[accion.ventana]->objeto(accion.objeto)->textoGrosor(accion.parametros[j].i);
        } break;
        case L_TEXTO:
        {
         _ventanas[accion.ventana]->objeto(accion.objeto)->texto(accion.parametros[j].s);
        } break;
        default : break;
       }
      }
     }
     catch(...)
     {
      cout << "Error, objeto inexistente." << endl;
     }
    }
    else
    {
     if(accion.retorno)
     {
      _datos[accion.ventana] = _ventanas[accion.ventana]->datos();
     }
     else
     {
      for(unsigned int j = 0;j < accion.parametros.size();j++)
      {
       switch(accion.parametros[j].t_parametro)
       {
        case L_X:
        {
         _ventanas[accion.ventana]->x(accion.parametros[j].i);
        } break;
        case L_Y:
        {
         _ventanas[accion.ventana]->y(accion.parametros[j].i);
        } break;
        case L_ANCHO:
        {
         _ventanas[accion.ventana]->ancho(accion.parametros[j].i);
        } break;
        case L_ALTO:
        {
         _ventanas[accion.ventana]->alto(accion.parametros[j].i);
        } break;
        case L_IMG:
        {
         _ventanas[accion.ventana]->textura(accion.parametros[j].i);
         _ventanas[accion.ventana]->set(V_TEXTURADO);
        } break;
        case L_VISIBLE:
        {
         if (accion.parametros[j].b)
          _ventanas[accion.ventana]->set(V_VISIBLE);
         else
          _ventanas[accion.ventana]->unset(V_VISIBLE);
        } break;
        case L_MOVIL:
        {
         if (accion.parametros[j].b)
          _ventanas[accion.ventana]->set(V_MOVIL);
         else
          _ventanas[accion.ventana]->unset(V_MOVIL);
        } break;
        case L_COLOR:
        {
         _ventanas[accion.ventana]->colorf(accion.parametros[j].f3);
        } break;
        case L_BCOLOR:
        {
         _ventanas[accion.ventana]->colorb(accion.parametros[j].f3);
        } break;
        default : break;
       }
      }
     }
    }
   }
  }
 }

 return res.first;
}

/*******************************************************************************
*
* Método ci_arriba.
* Recibe dos enteros con signo de 32 bits x e y.
* Cambia el estado de pulsación.
*
*******************************************************************************/
void c_sventana::ci_arriba(GLint x,GLint y)
{
 _RBE &= ~SV_CLICKABAJO;
}

/*******************************************************************************
*
* Método c_mover.
* Recibe dos flotantes x e y.
* Si hay alguna ventana con el foco, se está haciendo click y la ventan es móvil
* la mueve.
*
*******************************************************************************/
void c_sventana::c_mover(GLfloat x,GLfloat y)
{
 if((_RBE & SV_CLICKABAJO) && _id_foco && (_ventanas[_id_foco]->estado() & V_MOVIL))
 {
  _ventanas[_id_foco]->x(x - _x_foco);
  _ventanas[_id_foco]->y(y - _y_foco);
 }
}

/*******************************************************************************
*
* Método nuevo_objparametro.
* Recibe un nodo_parametro parametro.
* Interpreta el parámetro y lo configura en el objeto. Lanza el siguiente en
* caso de haberlo.
*
*******************************************************************************/
nodo_parametro * c_sventana::nuevo_objparametro(nodo_parametro * parametro)
{
 switch(parametro->t_parametro)
 {
  case L_ID: _c_oventana_id = parametro->i_valor; break;
  case L_X: _oventana->x(parametro->i_valor); break;
  case L_Y: _oventana->y(parametro->i_valor); break;
  case L_IMG: { } break;
  case L_ANCHO: _oventana->ancho(parametro->i_valor); break;
  case L_ALTO: _oventana->alto(parametro->i_valor); break;
  case L_COLOR: _oventana->colorf(parametro->fv3_valor); break;
  case L_BCOLOR: _oventana->colorb(parametro->fv3_valor);  break;
  case L_TCOLOR: _oventana->colort(parametro->fv3_valor); break;
  case L_FACTORTEXTO: _oventana->textoTamano(parametro->i_valor); break;
  case L_GROSORTEXTO: _oventana->textoGrosor(parametro->i_valor); break;
  case L_TEXTO: _oventana->texto(parametro->s_valor); break;
  default: {  } break;
 }
 if(parametro->parametro)
 {
  parametro->parametro = nuevo_objparametro(parametro->parametro);
 }
 delete parametro;
 return 0;
}

/*******************************************************************************
*
* Método nuevo_accparametro.
* Recibe un nodo_parametro parametro.
* Construye parámetros y los introduce en un vector. Si quedan más parámetros,
* hace la llamada apropiada.
*
*******************************************************************************/
nodo_parametro * c_sventana::nuevo_accparametro(nodo_parametro * parametro)
{
 c_parametro param;
 param.s = parametro->s_valor;
 param.f3[0] = parametro->fv3_valor[0];
 param.f3[1] = parametro->fv3_valor[1];
 param.f3[2] = parametro->fv3_valor[2];
 param.i = parametro->i_valor;
 param.t_parametro = parametro->t_parametro;
 param.b = parametro->b_valor;
 _eaparametros.push_back(param);
 if(parametro->parametro)
 {
  parametro->parametro = nuevo_accparametro(parametro->parametro);
 }
 delete parametro;
 return 0;
}

/*******************************************************************************
*
* Método nueva_accion.
* Recibe un nodo_accion accion.
* Construye la acción con los parámetros que procedan y, si hay más acciones,
* realiza la llamada apropiada.
*
*******************************************************************************/
nodo_accion * c_sventana::nueva_accion(nodo_accion * accion)
{
 _eaparametros.clear();
 c_accion ac;

 ac.ventana = accion->ventana;
 ac.objeto = accion->objeto;
 ac.retorno = accion->datos;

 if(!ac.retorno)
 {
  if(accion->parametro)
  {
   accion->parametro = nuevo_accparametro(accion->parametro);
  }
 }
 ac.parametros = _eaparametros;
 _oventana->nuevaaccion(_tevento,ac);
 if(accion->accion)
 {
  accion->accion = nueva_accion(accion->accion);
 }
 delete accion;
 return 0;
}

/*******************************************************************************
*
* Método nuevo_evento.
* Recibe un nodo_evento evento.
* Construye el evento y lanza el acciones de haberlas.
*
*******************************************************************************/
nodo_evento * c_sventana::nuevo_evento(nodo_evento * evento)
{
 _tevento = evento->t_evento;
 if(evento->accion)
 {
  evento->accion = nueva_accion(evento->accion);
 }
 if(evento->evento)
 {
  evento->evento = nuevo_evento(evento->evento);
 }
 delete evento;
 return 0;
}
/*******************************************************************************
*
* Método nuevo_objeto.
* Recibe un nodo_objeto objeto.
* Construye el objeto de ventana y lanza el constructor de parámetros en caso
* de haberlos.
*
*******************************************************************************/
nodo_objeto * c_sventana::nuevo_objeto(nodo_objeto * objeto)
{
 switch(objeto->t_objeto)
 {
  case L_OBOTON:
  {
   _oventana = new c_oboton();
  } break;
  case L_OTEXTO:
  {
   _oventana = new c_otexto();
  } break;
  case L_OCASILLERO:
  {
   _oventana = new c_ocasillero();
  } break;
  case L_OTEXTOEDITABLE:
  {
   _oventana = new c_otextoeditable();
  } break;
  case L_OSELECTOR:
  {
   _oventana = new c_oselector();
  } break;
  case L_OSELECTORFICHEROS:
  {
   _oventana = new c_oselectorficheros();
  } break;
  default: {} break;
 }
 _oventana->fuente(_fuente);
 if(objeto->parametro)
 {
  objeto->parametro = nuevo_objparametro(objeto->parametro);
 }
 if(objeto->evento)
 {
  objeto->evento = nuevo_evento(objeto->evento);
 }
 _oventana->chequear();
 _ventana->nobjeto(_c_oventana_id,_oventana);
 if(objeto->objeto)
 {
  objeto->objeto = nuevo_objeto(objeto->objeto);
 }
 delete objeto;
 return 0;
}

/*******************************************************************************
*
* Método nuevo_parametro.
* Recibe un nodo_parametro parametro.
* Configura la ventana con el parámetro parametro y lanza el siguiente en caso
* de haberlo. Después lanza el constructor de objetos.
*
*******************************************************************************/
nodo_parametro * c_sventana::nuevo_parametro(nodo_parametro * parametro)
{
 switch(parametro->t_parametro)
 {
  case L_ID: _c_ventana_id = parametro->i_valor; break;
  case L_X: _ventana->x(parametro->i_valor); break;
  case L_Y: _ventana->y(parametro->i_valor); break;
  case L_IMG: _ventana->textura(_texturas.at(parametro->i_valor)); _ventana->set(V_TEXTURADO); break;
  case L_ANCHO: _ventana->ancho(parametro->i_valor); break;
  case L_ALTO: _ventana->alto(parametro->i_valor); break;
  case L_VISIBLE:
  {
   if(parametro->b_valor) _ventana->set(V_VISIBLE);
   else _ventana->unset(V_VISIBLE);
  } break;
  case L_MOVIL:
  {
   if(parametro->b_valor) _ventana->set(V_MOVIL);
   else _ventana->unset(V_MOVIL);
  } break;
  case L_COLOR: _ventana->colorf(parametro->fv3_valor);  break;
  case L_BCOLOR: _ventana->colorb(parametro->fv3_valor); break;
  default: {  } break;
 }
 if(parametro->parametro)
 {
  parametro->parametro = nuevo_parametro(parametro->parametro);
 }
 delete parametro;
 return 0;
}

/*******************************************************************************
*
* Método nueva_ventana.
* Recibe un nodo_ventana ventana.
* Construye una ventana y lanza los configuradores de parámetros y de objetos.
*
*******************************************************************************/
nodo_ventana * c_sventana::nueva_ventana(nodo_ventana * ventana)
{
 if(ventana->parametro)
 {
  ventana->parametro = nuevo_parametro(ventana->parametro);
 }
 if(ventana->objeto)
 {
  ventana->objeto = nuevo_objeto(ventana->objeto);
 }
 delete ventana;
 return 0;
}

/*******************************************************************************
*
* Método generar_ventanas.
* Recibe un nodo_ventanas nventanas.
* Genera todas las ventanas a través de la jerarquía de nodos.
*
*******************************************************************************/
nodo_ventanas * c_sventana::generar_ventanas(nodo_ventanas * nventanas)
{
 if(nventanas->ventana)
 {
  _ventana = new c_ventana;
  nventanas->ventana = nueva_ventana(nventanas->ventana);
  _c_oventana_id = 0;
  _ventanas[_c_ventana_id] = _ventana;
  _vordenadas.push_back(_c_ventana_id);
  _c_ventana_id = 0;
 }
 if(nventanas->ventanas)
 {
  nventanas->ventanas = generar_ventanas(nventanas->ventanas);
 }
 delete nventanas;
 return 0;
}

/*******************************************************************************
*
* Método recogercaracter.
* Recibe un carácter c.
* Envía el carácter c a la ventana con el foco..
*
*******************************************************************************/
void c_sventana::recogercaracter(GLubyte c)
{
 if(_id_foco > 0)
 {
  _ventanas[_id_foco]->recogercaracter(c);
 }
}

/*******************************************************************************
*
* Método borrarcaracter.
* Borra un carácter de la ventana con el foco.
*
*******************************************************************************/
void c_sventana::borrarcaracter()
{
 if(_id_foco > 0)
 {
  _ventanas[_id_foco]->borrarcaracter();
 }
}

c_datos c_sventana::datos(GLuint i)
{
 if(_datos.count(i))
 {
  c_datos d = _datos[i];
  _datos.erase(i);
  return d;
 }
 svDatosNoExisten excepcion;
 throw excepcion;
}

/*******************************************************************************
*
* Método dibujar.
* Renderiza las ventanas.
*
*******************************************************************************/
void c_sventana::dibujar()
{
 glDisable(GL_DEPTH_TEST);
 GLfloat anchoLineaActual;
 glGetFloatv(GL_LINE_WIDTH,&anchoLineaActual);
 list<GLint>::reverse_iterator r = _vordenadas.rbegin(),fin = _vordenadas.rend();
 for(;r != fin;r++)
  _ventanas[(*r)]->dibujar();
 glLineWidth(anchoLineaActual);
 glEnable(GL_DEPTH_TEST);
}

